/*
	ProWesS, PROGS Windowing System

	Canvas object "CANV"

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 6 March 1995
	22.03.1996 adjusted for internal use of pixels
	26.03.1996 WindowSub is now set for all calls of client code which are
			   passed a CanvasInfo structure
	22.06.1996 fixed a problem with setting of canvas size
	19.09.1997 PW_CANVAS_ACTION_TIMEOUT added
	15.01.1998 PW_CANVAS_ACTION_CATCHSTART/CATCHEND/CATCH added
*/

#include <mem.h>
#include <win.h>

#include "ProWesS.h"
#include "PWTypeDef.h"
#include "PWHandler.h"

/* global statusses */
pt Xpix2pt, Ypix2pt;
pt Xpt2pix, Ypt2pix;

/* object definition */
typedef struct _CanvasObject {
	PWBaseObject;					/* common start for all objects */
	pt winxsiz, winysiz;			/* min size of canvas on screen */
	pt winxorg, winyorg;			/* origin of canvas on screen */
	Sprite *pointer;				/* pointer to use in window */
	Error (*Redraw)(struct _CanvasObject *, CanvasInfo *);	/* redraw action */
	Error (*Hit)(struct _CanvasObject *, CanvasInfo *); 	/* hit action */
	Error (*Do)(struct _CanvasObject *, CanvasInfo *);		/* do action */
	Error (*Exit)(struct _CanvasObject *, CanvasInfo *);	/* exit action */
	Error (*Move)(struct _CanvasObject *, CanvasInfo *);	/* move action */
	Error (*Scale)(struct _CanvasObject *, CanvasInfo *);	/* scale action */
	Error (*DragHit)(struct _CanvasObject *, CanvasInfo *); /* drag action */
	Error (*DragDo)(struct _CanvasObject *, CanvasInfo *);	/* drag action */
	Error (*DragEnd)(struct _CanvasObject *, CanvasInfo *); /* drag action */
	Error (*DragAdjust)(struct _CanvasObject *, CanvasInfo *, pt, pt);
	Error (*Time)(struct _CanvasObject *, CanvasInfo *);	/* timeout action */
	Error (*CatchStart)(struct _CanvasObject *, CanvasInfo *);
	Error (*CatchEnd)(struct _CanvasObject *, CanvasInfo *);
	Error (*Catch)(struct _CanvasObject *, CanvasInfo *, char key);
} CanvasObject;

/* object handling routines */

Error CanvasCreate(PWObject owner, PWObject *ret, PWEvent event, int *taglist)
{
	LinkPROforma();
	PWpixel2point(&Xpix2pt,&Ypix2pt);
	PWpoint2pixel(&Xpt2pix,&Ypt2pix);

	/* allocate specific object - which should be cleared (automatic) */
	return PWAllocate(event.region->system,sizeof(CanvasObject),ret);

	/* the events which have to be reacted upon are filled in by the */
	/* CanvasChange routine */
}

Error CanvasRemove(PWObject object)
{
	return PWRelease(object,object);
}

static void fillin(CanvasInfo *ri, PWSystem *system, PWRegion *reg, CanvasObject *obj)
{
	ri->xorg=obj->winxorg;
	ri->yorg=obj->winyorg;
	ri->gstate=system->gstate;
	ri->window=system->window;
	ri->xpixsiz=reg->hit_xend-reg->hit_xorg;
	ri->ypixsiz=reg->hit_yend-reg->hit_yorg;
	ri->xsiz=xpix2PF(ri->xpixsiz);
	ri->ysiz=ypix2PF(ri->ypixsiz);
	ri->xpixorg=reg->hit_xorg;
	ri->ypixorg=reg->hit_yorg;
	ri->xpixpos=system->ptr_xpos-system->win_xpos-reg->hit_xorg+system->xorg;
	ri->ypixpos=system->ptr_ypos-system->win_ypos-reg->hit_yorg+system->yorg;
	ri->xpos=xpix2PF(ri->xpixpos);
	ri->ypos=ypix2PF(ri->ypixpos);
	ri->Xpix2pt=Xpix2pt;
	ri->Ypix2pt=Ypix2pt;
	ri->Xpt2pix=Xpt2pix;
	ri->Ypt2pix=Ypt2pix;
	ri->zero=0;
}

Error CanvasDraw(PWObject object)
{
	Error err=ERR_OK;
	CanvasObject *obj=(CanvasObject *)object;
	PWRegion *reg=obj->event.region;
	PWSystem *system=obj->event.region->system;
	Gstate gstate=system->gstate;
	CanvasInfo rinfo;

	if (!gstate) return ERR_OK;

	/* call client defined canvas draw routine */
	if (obj->Redraw)
	{
		fillin(&rinfo,system,reg,obj);
		PFPWindowSubSet(gstate, rinfo.xpixsiz, rinfo.ypixsiz,
								reg->hit_xorg, reg->hit_yorg);
		err=obj->Redraw(obj,&rinfo);
		PFWindowSubRestore(gstate);
	}
	return err;
}

Error CanvasDetermineSize(PWObject object)
{
	CanvasObject *obj=(CanvasObject *)object;
	PWRegion *region=obj->event.region;

	region->hit_off_lft=region->hit_off_rgt=
	region->hit_off_top=region->hit_off_bot=0;

	region->tot_xsiz=obj->winxsiz;
	region->tot_ysiz=obj->winysiz;

	return ERR_OK;
}

Error CanvasEntry(PWObject object)
{
	CanvasObject *obj=(CanvasObject *)object;
	Error err=ERR_OK;

	if (obj->pointer) PWChange((PWObject)obj, PW_POINTER, obj->pointer, NULL);

	if (obj->Move)
	{
		CanvasInfo rinfo;
		PWRegion *reg=obj->event.region;
		PWSystem *system=reg->system;
		Gstate gstate=system->gstate;
		
		fillin(&rinfo,system,reg,obj);
		PFPWindowSubSet(gstate, rinfo.xpixsiz, rinfo.ypixsiz,
								reg->hit_xorg, reg->hit_yorg);
		err=obj->Move(obj,&rinfo);
		PFWindowSubRestore(gstate);
	}
	return err;
}

Error CanvasHandle(PWObject object, PWWait event)
{
	Error err=ERR_OK;
	CanvasObject *obj=(CanvasObject *)object;
	PWRegion *reg=obj->event.region;
	PWSystem *system=reg->system;
	Gstate gstate=system->gstate;
	CanvasInfo rinfo;

	fillin(&rinfo,system,reg,obj);
	PFPWindowSubSet(gstate, rinfo.xpixsiz, rinfo.ypixsiz,
								reg->hit_xorg, reg->hit_yorg);

	switch (event)
	{
	case PW_EVENT_HIT:
		err=obj->Hit(obj,&rinfo);
		break;
	case PW_EVENT_DO:
		err=obj->Do(obj,&rinfo);
		break;
	case PW_EVENT_MOVE:
		err=obj->Move(obj,&rinfo);
		break;
	case PW_EVENT_DRAGEND:
		if (obj->DragEnd)
			err=obj->DragEnd(obj,&rinfo);
		break;
	case PW_EVENT_HITDRAG:
		err=obj->DragHit(obj,&rinfo);
		break;
	case PW_EVENT_DODRAG:
		err=obj->DragDo(obj,&rinfo);
		break;
	case PW_EVENT_DRAGADJUST:
		err=obj->DragAdjust(obj,&rinfo,xpix2PF(system->Xdragadjust),
									   xpix2PF(system->Ydragadjust));
		break;
	case PW_EVENT_TIMEOUT:
		if (obj->Time)
			err=obj->Time(obj,&rinfo);
		break;
	case PW_EVENT_CATCHSTART:
		if (obj->CatchStart)
			err=obj->CatchStart(obj,&rinfo);
		break;
	case PW_EVENT_CATCHEND:
		if (obj->CatchEnd)
			err=obj->CatchEnd(obj,&rinfo);
		break;
	case PW_EVENT_CATCH:
		if (obj->Catch)
			err=obj->Catch(obj,&rinfo,system->catch);
		break;
	}
	PFWindowSubRestore(gstate);
	return err;
}

Error CanvasExit(PWObject object)
{
	Error err=ERR_OK;
	CanvasObject *obj=(CanvasObject *)object;

	if (obj->Exit)
	{
		PWRegion *reg=obj->event.region;
		PWSystem *system=reg->system;
		Gstate gstate=system->gstate;
		CanvasInfo rinfo;

		fillin(&rinfo,system,reg,obj);
		PFPWindowSubSet(gstate, rinfo.xpixsiz, rinfo.ypixsiz,
								reg->hit_xorg, reg->hit_yorg);
		err=obj->Exit(obj,&rinfo);
		PFWindowSubRestore(gstate);
		err=obj->Exit(obj,&rinfo);
	}
	return err;
}

Error DoScale(PWObject object, pt xinc, pt yinc)
{
	Error err=ERR_OK;
	CanvasObject *obj=(CanvasObject *)object;

	if (obj->Scale)
	{
		PWRegion *reg=obj->event.region;
		PWSystem *system=reg->system;
		Gstate gstate=system->gstate;
		CanvasInfo rinfo;

		fillin(&rinfo,system,reg,obj);
		PFPWindowSubSet(gstate, rinfo.xpixsiz, rinfo.ypixsiz,
								reg->hit_xorg, reg->hit_yorg);
		err=obj->Scale(obj,&rinfo);
		PFWindowSubRestore(gstate);
	}
	return err;
}

Error CanvasChange(PWObject object, int *taglist)
{
	CanvasObject *obj=(CanvasObject *)object;
	PWSystem *system=obj->event.region->system;
	Error err=ERR_OK;
	PWWait *wait=&obj->event.region->wait_event;
	char unrecognised_tag=FALSE;

	while (*taglist)
	{
		switch (*taglist)
		{
		case PW_CANVAS_POINTER:
			obj->pointer=(Sprite *)(*(taglist+1));
			break;
		case PW_CANVAS_ACTION_REDRAW:
			obj->Redraw=(Error (*)(CanvasObject *,CanvasInfo *))(*(taglist+1));
			break;
		case PW_CANVAS_ACTION_SCALE:
			obj->Scale=(Error (*)(CanvasObject *,CanvasInfo *))(*(taglist+1));
			break;
		case PW_CANVAS_ACTION_EXIT:
			obj->Exit=(Error (*)(CanvasObject *,CanvasInfo *))(*(taglist+1));
			break;
		case PW_CANVAS_ACTION_HIT:
			obj->Hit=(Error (*)(CanvasObject *,CanvasInfo *))(*(taglist+1));
			if (obj->Hit) *wait|= PW_EVENT_HIT;
					 else *wait&=~PW_EVENT_HIT;
			break;
		case PW_CANVAS_ACTION_DO:
			obj->Do=(Error (*)(CanvasObject *,CanvasInfo *))(*(taglist+1));
			if (obj->Do) *wait|= PW_EVENT_DO;
					else *wait&=~PW_EVENT_DO;
			break;
		case PW_CANVAS_ACTION_MOVE:
			obj->Move=(Error (*)(CanvasObject *,CanvasInfo *))(*(taglist+1));
			if (obj->Move) *wait|= PW_EVENT_MOVE;
					  else *wait&=~PW_EVENT_MOVE;
			break;
		case PW_CANVAS_ACTION_TIMEOUT:
			obj->Time=(Error (*)(CanvasObject *,CanvasInfo *))(*(taglist+1));
			if (obj->Time) *wait|= PW_EVENT_TIMEOUT;
					  else *wait&=~PW_EVENT_TIMEOUT;
			break;
		case PW_CANVAS_ACTION_HITDRAG:
			obj->DragHit=(Error (*)(CanvasObject *,CanvasInfo *))(*(taglist+1));
			if (obj->DragHit) *wait|= PW_EVENT_HITDRAG;
						 else *wait&=~PW_EVENT_HITDRAG;
			break;
		case PW_CANVAS_ACTION_DODRAG:
			obj->DragDo=(Error (*)(CanvasObject *,CanvasInfo *))(*(taglist+1));
			if (obj->DragDo) *wait|= PW_EVENT_DODRAG;
						else *wait&=~PW_EVENT_DODRAG;
			break;
		case PW_CANVAS_ACTION_DRAGEND:
			obj->DragEnd=(Error (*)(CanvasObject *,CanvasInfo *))(*(taglist+1));
			break;
		case PW_CANVAS_ACTION_DRAGADJUST:
			obj->DragAdjust=(Error (*)(CanvasObject *,CanvasInfo *,pt,pt))(*(taglist+1));
			if (obj->DragDo) *wait|= PW_EVENT_DRAGADJUST;
						else *wait&=~PW_EVENT_DRAGADJUST;
			break;

		case PW_CANVAS_TIMEOUT:
			obj->event.region->timeout=*(taglist+1);
			break;

		case PW_CANVAS_SIZE_PIX:
			system->status|=SSTATUS_WINCHANGE;
			obj->winxsiz=short2pt(*(taglist+1));
			obj->winysiz=short2pt(*(taglist+2));
			break;
		case PW_CANVAS_XSIZE_PIX:
			system->status|=SSTATUS_WINCHANGE;
			obj->winxsiz=short2pt(*(taglist+1));
			break;
		case PW_CANVAS_YSIZE_PIX:
			system->status|=SSTATUS_WINCHANGE;
			obj->winysiz=short2pt(*(taglist+1));
			break;
		case PW_CANVAS_SIZE:
			system->status|=SSTATUS_WINCHANGE;
			obj->winxsiz=fixmul(*(taglist+1),Xpt2pix);
			obj->winysiz=fixmul(*(taglist+2),Ypt2pix);
			break;
		case PW_CANVAS_XSIZE:
			system->status|=SSTATUS_WINCHANGE;
			obj->winxsiz=fixmul(*(taglist+1),Xpt2pix);
			break;
		case PW_CANVAS_YSIZE:
			system->status|=SSTATUS_WINCHANGE;
			obj->winysiz=fixmul(*(taglist+1),Ypt2pix);
			break;
		case PW_CANVAS_ORIGIN:
			obj->winxorg=*(taglist+1);
			obj->winyorg=*(taglist+2);
			system->status|=SSTATUS_REDRAW;
			obj->status|=OSTATUS_REDRAW;
			break;
		case PW_CANVAS_XORIGIN:
			obj->winxorg=*(taglist+1);
			system->status|=SSTATUS_REDRAW;
			obj->status|=OSTATUS_REDRAW;
			break;
		case PW_CANVAS_YORIGIN:
			obj->winyorg=*(taglist+1);
			system->status|=SSTATUS_REDRAW;
			obj->status|=OSTATUS_REDRAW;
			break;
		case PW_CANVAS_XSCROLL:
			obj->winxorg+=*(taglist+1);
			system->status|=SSTATUS_REDRAW;
			obj->status|=OSTATUS_REDRAW;
			break;
		case PW_CANVAS_YSCROLL:
			obj->winyorg+=*(taglist+1);
			system->status|=SSTATUS_REDRAW;
			obj->status|=OSTATUS_REDRAW;
			break;

		case PW_CANVAS_REDRAW:
			system->status|=SSTATUS_REDRAW;
			obj->status|=OSTATUS_REDRAW;
			break;

		case PW_CANVAS_ACTION_CATCH:
			obj->Catch=(Error (*)(CanvasObject *,CanvasInfo *,char))(*(taglist+1));
			if (obj->Catch) *wait|= PW_EVENT_CATCH;
					   else *wait&=~PW_EVENT_CATCH;
			break;
		case PW_CANVAS_ACTION_CATCHSTART:
			obj->CatchStart=(Error (*)(CanvasObject *,CanvasInfo *))(*(taglist+1));
			if (obj->CatchStart) *wait|= PW_EVENT_CATCHSTART;
							else *wait&=~PW_EVENT_CATCHSTART;
			break;
		case PW_CANVAS_ACTION_CATCHEND:
			obj->CatchEnd=(Error (*)(CanvasObject *,CanvasInfo *))(*(taglist+1));
			if (obj->CatchEnd) *wait|= PW_EVENT_CATCHEND;
						  else *wait&=~PW_EVENT_CATCHEND;
			break;

		default :
			if (*taglist&(~PW_TAGMASK_SKIP))
				unrecognised_tag=TRUE;
			break;
		}
		taglist=PWSkipTag(taglist);
	}
	if (unrecognised_tag) err=ERR_IPAR;
	return err;
}

Error Query(PWObject object, int what, void *ret)
{
	Error err=ERR_OK;
	CanvasObject *obj=(CanvasObject *)object;
	PWRegion *reg=obj->event.region;

	switch (what)
	{
	case PW_CANVAS_XORIGIN:
		*((pt *)ret)=obj->winxorg;
		break;
	case PW_CANVAS_YORIGIN:
		*((pt *)ret)=obj->winyorg;
		break;
	case PW_CANVAS_XSIZE_PIX:
		/* this code could say : */
		/* *((pt *)ret)=reg->hit_xend-reg->hit_xorg; */
		/* but this doesn't work when called from inside a PWScale event */
		/* As a canvas doesn't have a margin anyway, use (less accurate) */
		*((pt *)ret)=pt2rshort(reg->tot_xsiz);
		break;
	case PW_CANVAS_YSIZE_PIX:
		/* this code could say : */
		/* *((pt *)ret)=reg->hit_yend-reg->hit_yorg; */
		/* but this doesn't work when called from inside a PWScale event */
		/* As a canvas doesn't have a margin anyway, use (less accurate) */
		*((pt *)ret)=pt2rshort(reg->tot_ysiz);
		break;
	case PW_CANVAS_XSIZE:
		/* this code could say : */
		/* *((pt *)ret)=reg->hit_xend-reg->hit_xorg; */
		/* but this doesn't work when called from inside a PWScale event */
		/* As a canvas doesn't have a margin anyway, use (less accurate) */
		*((pt *)ret)=fixmul(reg->tot_xsiz,Xpix2pt);
		break;
	case PW_CANVAS_YSIZE:
		/* this code could say : */
		/* *((pt *)ret)=reg->hit_yend-reg->hit_yorg; */
		/* but this doesn't work when called from inside a PWScale event */
		/* As a canvas doesn't have a margin anyway, use (less accurate) */
		*((pt *)ret)=fixmul(reg->tot_ysiz,Ypix2pt);
		break;
	default:
		err=ERR_IPAR;
		break;
	}
	return err;
}

PWTypeDef init={
	PW_TYPEDEF_FLAG, PW_TYPE_CANVAS, PW_TYPE_EVENT_REGION,
	&CanvasCreate, &CanvasChange, &Query, &CanvasRemove,
	&CanvasEntry, &CanvasExit, &CanvasHandle,
	NULL, &CanvasDraw, &CanvasDetermineSize, &DoScale
};
